type movie = string * string * float * int;;
type studio_gross = string * float;;

(* we don't care about the movie names or release years *)
let get (m : movie) : studio_gross =
  let (_,s,g,_) = m in (s,g);;

(* compare by studio name only *)
let studiosort (m1 : movie) (m2 : movie) : int =
  let (_,s1,_,_) = m1 in
  let (_,s2,_,_) = m2 in
  compare s1 s2
;;

let data4_list : movie list = [
  ("MA","S1",1.0,2000);
  ("MB","S2",2.1,2001);
  ("MC","S1",3.2,2002);
  ("MD","S3",4.3,2003);
  ("ME","S2",5.4,2004)
];;

(* here's a reasonable implementation of the rec version
   of by_studio from Assignment 2 that doesn't use sort *)
let by_studio_rec (movies:movie list) : sg list =
  let rec addin (s, g) (sgs : sg list) : sg list =
    match sgs with
      [] -> [(s, g)]
    | (s',g')::rest ->
       if s' = s
       then (s',g'+.g)::rest
       else (s',g')::addin (s,g) rest
  in
  let rec all_movies (movies:movie list) (sgs:sg list) : sg list =
    match movies with
      [] -> sgs
    | m::rest -> all_movies rest (addin (get m) sgs)
  in
  all_movies movies []
;;

let bsrec4 = by_studio_rec data4_list ;;

(* and here is a similar version with reduce and sort *)
let by_studio_red (movies:movie list) : sg list =
  let addin (acc:sg list) (m:movie) : sg list =
    let (s,g) = get m in
    match acc with
    | [] -> [(s,g)]
    | (hs,hg)::tl ->
       if s = hs then (s,g+.hg)::tl
       else (s,g)::acc
  in
  List.fold_left addin [] (List.sort studiosort movies)
;;

let bsred4 = by_studio_red data4_list ;;

(* Now, let's do the same thing with parallel sequences,
   using the library we'll use in the assignment! *)

#mod_use "sequence.ml"
module S = Sequence.S


let data4 : movie S.t =
  S.cons ("MA","S1",1.0,2000)
   (S.cons ("MB","S2",2.1,2001)
     (S.cons ("MC","S1",3.2,2002)
       (S.cons ("MD","S3",4.3,2003)
         (S.cons ("ME","S2",5.4,2004)
           (S.empty ())))))

(* These functions are just for testing purposes,
   you CANNOT use them in the assignment since they are not parallel at all *)
let seq_to_list (s : 'a S.t) : 'a list =
  Array.to_list (S.array_of_seq s)
let seq_of_list (l : 'a list) : 'a S.t =
 S.seq_of_array (Array.of_list l)

(* our List based implementations of filter and sort
   note that in the assignment we will ask you to implement these in parallel,
   so you cannot convert them to lists. *)
let seq_filter (f : 'a -> bool) (s : 'a S.t) : 'a S.t =
  seq_of_list (List.filter f (seq_to_list s))

(* One of the lectures talks about sorting sequences in parallel! *)
let seq_sort (compare : 'a -> 'a -> int) (s : 'a S.t) : 'a S.t =
  seq_of_list (List.sort compare (seq_to_list s))

(* test to see if our bogus sort function works *)
let _ = assert ([1;2;3] = seq_to_list (seq_sort compare (seq_of_list [2;3;1])))

(*
  look back at A2 for by_studio done with recursion
  (warning: you might cringe at your old code now
  that you're a seasoned OCamlist). You could also
  do that "rec function" version with a fold now!
  But the design with only parallel operations takes
  perhaps a bit more thought?
*)

type status =
  | Start
     (* to start scanning with *)
  | ChangedStudio of studio_gross * studio_gross
     (* the last value of the previous studio and a new value *)
  | SameStudio of studio_gross
     (* keep counting in the same studio *)

(* Start has to be unit *)
(* for all s : status, aux Start s = s *)
(* for all s : status, aux s Start = s *)

(* aux should try to be associative *)
(* for all s1 s2 s3 : status, aux s1 (aux s2 s3) = aux (aux s1 s2) s3 *)
(* Now, this is not true for our aux but the way we use the type
   doesn't seem to break the parallel functions  *)

let by_studio_seq (movies : movie S.t) : studio_gross S.t =

  (* Sort by studio to ensure that movies from the same studio are consecutive  *)
  let sorted : movie S.t = seq_sort studiosort movies in
  (* We want to use scan which only contains one type for the function and the elements, so let's map *)
  let mapped : status S.t = S.map (fun m -> SameStudio (get m)) sorted in
  (* Our strategy is to use the change points in the sequence as the studio aggregates,
     but there is no change point for the last studio, so let's add a dummy one *)
  let added_dummy : status S.t = S.append mapped (S.singleton (SameStudio ("", nan))) in

  (* accumulate like studios together,
     stash the final result away when complete *)
  let aux (h1 : status) (h2 : status) : status =
    match h1 , h2 with
    | Start , _ -> h2
    | _ , Start -> h1
    | SameStudio (s1, g1) , SameStudio (s2, g2) ->
      if s1 = s2 then SameStudio (s1, g1 +. g2) else ChangedStudio ((s1, g1), (s2, g2))
    | ChangedStudio (_, (s1, g1)) , SameStudio (s2, g2) ->
      if s1 = s2 then SameStudio (s1, g1 +. g2) else ChangedStudio ((s1, g1), (s2, g2))
    | _ , ChangedStudio _ -> h2
  in
  let scanned : status S.t = S.scan aux Start added_dummy in

  (* helper function: did we stash something here? *)
  let is_changed (s : status) : bool =
    match s with
    | ChangedStudio _ -> true
    | _ -> false
  in

  (* grab only the items that store saved final studio results *)
  let filtered : status S.t = seq_filter is_changed scanned in
  (* grab only the saved final studio results themselves*)
  let extract (s : status) : studio_gross =
    match s with
    | ChangedStudio (sg, _) -> sg
    | _ -> failwith "Can't extract"
  in
  S.map extract filtered
;;

let bsseq4 = by_studio_seq data4;;

(* convert to list just to be able to inspect the result: *)
let _ = seq_to_list bsseq4 ;;
